/*
 * JBoss, Home of Professional Open Source
 * Copyright 2007, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags.
 * See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions
 * of the GNU Lesser General Public License, v. 2.1.
 * This program is distributed in the hope that it will be useful, but WITHOUT A
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License,
 * v.2.1 along with this distribution; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * (C) 2007,
 * @author JBoss Inc.
 */
package org.jboss.wst11.xts.impl;

import com.arjuna.ats.arjuna.common.Uid;
import com.arjuna.ats.jta.transaction.Transaction;
import com.arjuna.ats.jta.xa.XidImple;
import com.arjuna.mw.wst.TxContext;
import com.arjuna.mw.wst11.TransactionManagerFactory;
import com.arjuna.mw.wst11.UserTransactionFactory;
import com.arjuna.mwlabs.wst11.at.ContextImple;
import com.arjuna.mwlabs.wst11.at.context.TxContextImple;
import com.arjuna.webservices.SoapFault;
import com.arjuna.webservices11.wsat.AtomicTransactionConstants;
import com.arjuna.wsc.AlreadyRegisteredException;
import com.arjuna.wsc.InvalidCreateParametersException;
import com.arjuna.wsc.InvalidProtocolException;
import com.arjuna.wsc.InvalidStateException;
import com.arjuna.wsc.NoActivityException;
import com.arjuna.wst.UnknownTransactionException;
import com.arjuna.wst.WrongStateException;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import javax.transaction.NotSupportedException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.TransactionManager;
import javax.transaction.xa.XAException;
import javax.transaction.xa.Xid;
import org.apache.log4j.Logger;
import org.jboss.tm.TransactionManagerLocator;
import org.oasis_open.docs.ws_tx.wscoor._2006._06.CoordinationContext;
import org.oasis_open.docs.ws_tx.wscoor._2006._06.CoordinationContextType;
import com.arjuna.mwlabs.wst11.at.ContextFactoryImple;
import com.arjuna.wsc11.ContextFactory;
import com.arjuna.wsc11.ContextFactoryMapper;
import com.arjuna.wst.Durable2PCParticipant;
import org.jboss.wst.xts.WSTransactionBridge;
import org.jboss.wst.xts.WSTransactionBridgeManager;
import org.oasis_open.docs.ws_tx.wscoor._2006._06.Expires;

/**
 *
 * @author wklaczynski@sabaservice.pl, 2007-04-30
 */
public abstract class AbstractTransactionBridgeManagerImpl implements WSTransactionBridgeManager {

    private static Logger log = Logger.getLogger(AbstractTransactionBridgeManagerImpl.class);
    private static ConcurrentMap<String, Transaction> txIdMappings = new ConcurrentHashMap<String, Transaction>();
    private static ConcurrentMap<Transaction, String> xidMappings = new ConcurrentHashMap<Transaction, String>();
    private static ConcurrentMap<Transaction, WSTransactionBridge> txJtaSubBridgeMappings = new ConcurrentHashMap<Transaction, WSTransactionBridge>();
    private static ConcurrentMap<String, WSTransactionBridge> txXtsSubBridgeMappings = new ConcurrentHashMap<String, WSTransactionBridge>();

    /**
     *
     */
    public AbstractTransactionBridgeManagerImpl() {
    }

    /**
     *
     * @return Service transaction bridge
     * @throws javax.transaction.SystemException
     */
    @Override
    public WSTransactionBridge getServiceTransactionBridge() throws SystemException {
        try {
            TxContextImple xtransaction = (TxContextImple) TransactionManagerFactory.transactionManager().currentTransaction();
            if (xtransaction != null) {

                String txId = xtransaction.identifier();
                Transaction transaction = txIdMappings.get(txId);
                if (transaction == null) {
                    final CoordinationContextType coordinationContext = xtransaction.context().getCoordinationContext();
                    Expires expiresInstance = coordinationContext.getExpires();
                    final long transactionExpires = (expiresInstance != null ? expiresInstance.getValue() : 0);
                    transaction = createServiceMappings(txId, transactionExpires);
                }
                return txJtaSubBridgeMappings.get(transaction);
            } else {
                return new EmptyTransactionBridgeImpl();

            }
        } catch (Exception ex) {
            throw new SystemException(ex.getMessage());
        }
    }

    /**
     *
     * @return Client transaction bridge
     * @throws javax.transaction.SystemException
     */
    @Override
    public WSTransactionBridge getClientTransactionBridge() throws SystemException {
        try {
            TransactionManager manager = TransactionManagerLocator.locateTransactionManager(true);
            if (manager.getStatus() == Status.STATUS_NO_TRANSACTION) {
                return new EmptyTransactionBridgeImpl();
            } else {
                if (manager.getTransaction() instanceof Transaction) {
                    Transaction transaction = (Transaction) manager.getTransaction();
                    String txId = xidMappings.get(transaction);
                    if (txId == null) {
                        txId = createClientMappings(transaction);
                    }
                    return txXtsSubBridgeMappings.get(txId);
                } else {
                    throw new NotSupportedException("Not supported non JBossTS transaction type");
                }
            }
        } catch (Exception ex) {
            throw new SystemException(ex.getMessage());
        }
    }

    private synchronized Transaction createServiceMappings(String txId, long timeout) throws XAException, SystemException, com.arjuna.wst.WrongStateException, com.arjuna.wst.UnknownTransactionException, AlreadyRegisteredException, com.arjuna.wst.SystemException {
        if (txIdMappings.containsKey(txId)) {
            return txIdMappings.get(txId);
        }
        BridgeServiceCleanup cleanup = new BridgeServiceCleanup(txId);
        TransactionManagerFactory.transactionManager().enlistForVolatileTwoPhase(cleanup, new Uid().toString() + "_V");

        Xid xid = new XidImple(new Uid());
        Transaction transaction = getSubordinateTransaction(xid, timeout);

        txIdMappings.put(txId, transaction);
        txJtaSubBridgeMappings.put(transaction, new ServiceTransactionBridgeImpl(transaction));

        log.info(String.format("Create XTS/JTA bridge from TxContextImple<WS-T 1.1 %s> to %s",txId,transaction));

        Durable2PCParticipant durableBridgeParticipantAT = getBridgeParticipantAT(xid);
        TransactionManagerFactory.transactionManager().enlistForDurableTwoPhase(durableBridgeParticipantAT, new Uid().toString() + "_D");
        return transaction;
    }

    private synchronized String createClientMappings(Transaction transaction) throws com.arjuna.wst.SystemException, SystemException, WrongStateException, UnknownTransactionException, InvalidCreateParametersException, SoapFault, IllegalStateException, RollbackException, InvalidProtocolException, InvalidStateException, NoActivityException, com.arjuna.mw.wst.exceptions.SystemException, com.arjuna.mw.wsas.exceptions.NoActivityException, com.arjuna.mw.wsas.exceptions.SystemException, com.arjuna.mw.wsas.exceptions.WrongStateException, Exception {
        if (xidMappings.containsKey(transaction)) {
            return xidMappings.get(transaction);
        }
        TransactionManager manager = TransactionManagerLocator.locateTransactionManager(true);

        TxContext currentTransaction = TransactionManagerFactory.transactionManager().currentTransaction();
        TxContextImple xtransaction = null;

        if (currentTransaction != null) {
            String curTxId = UserTransactionFactory.userTransaction().toString();
            if (!transaction.equals(txIdMappings.get(curTxId))) {
                xtransaction = createSubordinateTransaction(transaction.getTimeout(), currentTransaction);
            }
        }

        BridgeClientCleanup cleanup = new BridgeClientCleanup(transaction);
        manager.getTransaction().registerSynchronization(cleanup);


        if (xtransaction == null) {
            BridgeXAResourceAT xaBridge = new BridgeXAResourceAT();
            manager.getTransaction().enlistResource(xaBridge);
            xtransaction = xaBridge.getTransactionContext();
        }

        ClientTransactionBridgeImpl impl = new ClientTransactionBridgeImpl(xtransaction);
        String txId = xtransaction.identifier();

        log.info(String.format("Create JTA/XTS bridge from %s to TxContextImple<WS-T 1.1 %s>",transaction,txId));

        xidMappings.put(transaction, txId);
        txXtsSubBridgeMappings.put(txId, impl);

        return txId;
    }

    /**
     *
     * @param xid
     * @param timeout
     * @return New or imported subordinate transaction
     * @throws javax.transaction.xa.XAException
     */
    protected abstract Transaction getSubordinateTransaction(Xid xid, long timeout) throws XAException;

    /**
     *
     * @param xid
     * @return New created bridge participant
     */
    protected abstract Durable2PCParticipant getBridgeParticipantAT(Xid xid);

    /**
     *
     * @param txId
     */
    public static void disassociateServiceBridge(String txId) {
        Transaction transaction = txIdMappings.get(txId);
        if (transaction != null) {
            txJtaSubBridgeMappings.remove(transaction);
        }
        txIdMappings.remove(txId);
    }

    /**
     *
     * @param transaction
     */
    public static void disassociateClientBridge(Transaction transaction) {
        String txId = xidMappings.get(transaction);
        if (txId != null) {
            txXtsSubBridgeMappings.remove(transaction);
        }
        xidMappings.remove(txId);
    }

    /**
     *
     * @param timeout
     * @param currentTransaction
     * @return New created XTS subordinate transaction
     * @throws com.arjuna.wsc.NoActivityException
     * @throws com.arjuna.wsc.InvalidProtocolException
     * @throws javax.transaction.SystemException
     * @throws com.arjuna.mw.wsas.exceptions.NoActivityException
     * @throws com.arjuna.mw.wsas.exceptions.SystemException
     * @throws javax.transaction.RollbackException
     * @throws com.arjuna.wsc.InvalidCreateParametersException
     */
    public static TxContextImple createSubordinateTransaction(long timeout, TxContext currentTransaction) throws NoActivityException, InvalidProtocolException, SystemException, com.arjuna.mw.wsas.exceptions.NoActivityException, com.arjuna.mw.wsas.exceptions.SystemException, RollbackException, InvalidCreateParametersException {
        final ContextFactoryMapper contextFactoryMapper = ContextFactoryMapper.getFactory();
        final ContextFactory contextFactory = contextFactoryMapper.getContextFactory(AtomicTransactionConstants.WSAT_PROTOCOL);
        final ContextFactoryImple imple = (ContextFactoryImple) contextFactory;
        Long expires = (timeout > 0 ? new Long(timeout) : null);
        CoordinationContext currentContext = currentCoordinationContext(currentTransaction);
        CoordinationContextType coordinationContext = imple.create(AtomicTransactionConstants.WSAT_PROTOCOL, expires, currentContext);
        TxContextImple transaction = new TxContextImple(new ContextImple(coordinationContext));
        return transaction;
    }

    private static CoordinationContext currentCoordinationContext(TxContext context) throws SystemException {
        final TxContextImple txContext = (TxContextImple) context;
        final CoordinationContextType savedContext = txContext.context().getCoordinationContext();
        final CoordinationContext current = new CoordinationContext();
        current.setCoordinationType(savedContext.getCoordinationType());
        current.setExpires(savedContext.getExpires());
        current.setIdentifier(savedContext.getIdentifier());
        current.setRegistrationService(savedContext.getRegistrationService());
        return current;
    }
}