/*
 * JBoss, Home of Professional Open Source
 * Copyright 2007, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags.
 * See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions
 * of the GNU Lesser General Public License, v. 2.1.
 * This program is distributed in the hope that it will be useful, but WITHOUT A
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License,
 * v.2.1 along with this distribution; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *
 * (C) 2007,
 * @author JBoss Inc.
 */
package org.jboss.wst11.xts.impl;

import com.arjuna.ats.arjuna.coordinator.TwoPhaseOutcome;
import com.arjuna.mw.wsas.exceptions.NoActivityException;
import com.arjuna.mw.wsas.exceptions.SystemException;
import com.arjuna.mwlabs.wscf.model.twophase.arjunacore.subordinate.SubordinateCoordinator;
import com.arjuna.mwlabs.wst11.at.ContextFactoryImple;
import com.arjuna.mwlabs.wst11.at.ContextImple;
import com.arjuna.mwlabs.wst11.at.context.TxContextImple;
import com.arjuna.webservices11.ServiceRegistry;
import com.arjuna.webservices11.wsarj.InstanceIdentifier;
import com.arjuna.webservices11.wsat.AtomicTransactionConstants;
import com.arjuna.webservices11.wscoor.CoordinationConstants;
import com.arjuna.wsc.InvalidProtocolException;
import com.arjuna.wsc11.ContextFactory;
import com.arjuna.wsc11.ContextFactoryMapper;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.HashSet;
import java.util.Set;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;
import javax.xml.ws.wsaddressing.W3CEndpointReference;
import javax.xml.ws.wsaddressing.W3CEndpointReferenceBuilder;
import org.apache.log4j.Logger;
import org.jboss.jbossts.xts.recovery.participant.at.XTSATRecoveryManager;
import org.oasis_open.docs.ws_tx.wscoor._2006._06.CoordinationContext;
import org.oasis_open.docs.ws_tx.wscoor._2006._06.CoordinationContextType;
import org.oasis_open.docs.ws_tx.wscoor._2006._06.Expires;

/**
 *
 * @author wklaczynski@sabaservice.pl, 2007-04-30
 */
public class BridgeXAResourceAT implements XAResource, Serializable {

    private static final long serialVersionUID = 7526481155678776147L;
    private static Logger log = Logger.getLogger(BridgeXAResourceAT.class);
    private Xid currentXid;
    private Xid prepared;
    private Xid committed;
    private Xid rolledback;
    private int transactionTimeoutSeconds;
    private final Set knownXids = new HashSet();
    private final Set successfulXids = new HashSet();
    private String coordinatorId;
    transient private boolean recovered;
    transient private SubordinateCoordinator coordinator;
    transient private TxContextImple context;

    /**
     *
     */
    public BridgeXAResourceAT() {
        this.coordinator = null;
        this.coordinatorId = null;
        this.recovered = false;
    }

    /**
     *
     * @param xid
     * @param flags
     * @throws javax.transaction.xa.XAException
     */
    public void start(Xid xid, int flags) throws XAException {
        if (xid == null) {
            throw new XAException(XAException.XAER_INVAL);
        }
        if (currentXid != null) {
            throw new XAException(XAException.XAER_OUTSIDE);
        }
        if (flags != XAResource.TMNOFLAGS && flags != XAResource.TMJOIN && flags != XAResource.TMRESUME) {
            throw new XAException(XAException.XAER_INVAL);
        }
        if (flags == XAResource.TMNOFLAGS || flags == XAResource.TMJOIN) {
            if (knownXids.contains(xid)) {
                throw new XAException(XAException.XAER_DUPID);
            }
            try {
                knownXids.add(xid);
                getCoordinator(xid);
            } catch (Exception ex) {
                throw new XAException(XAException.XAER_RMFAIL);
            }
        }
        if (flags == XAResource.TMRESUME) {
            if (!knownXids.contains(xid)) {
                throw new XAException(XAException.XAER_NOTA);
            }
        }
        currentXid = xid;
    }

    /**
     *
     * @param xid
     * @param flags
     * @throws javax.transaction.xa.XAException
     */
    public void end(Xid xid, int flags) throws XAException {
        if (xid == null) {
            throw new XAException(XAException.XAER_INVAL);
        }
        if (!knownXids.contains(xid)) {
            throw new XAException(XAException.XAER_NOTA);
        }
        if (flags != XAResource.TMSUSPEND && flags != XAResource.TMSUCCESS && flags != XAResource.TMFAIL) {
            throw new XAException(XAException.XAER_INVAL);
        }
        if (flags == XAResource.TMSUSPEND) {
            currentXid = null;
        }
        if (flags == XAResource.TMSUCCESS) {
            coordinator.end(recovered);
            successfulXids.add(xid);
            if (xid.equals(currentXid)) {
                currentXid = null;
            }
        }
        if (flags == XAResource.TMFAIL) {
            if (xid.equals(currentXid)) {
                currentXid = null;
            }
        }

    }

    /**
     *
     * @param xid
     * @return Prepare result
     * @throws javax.transaction.xa.XAException
     */
    public int prepare(Xid xid) throws XAException {
        if (xid == null) {
            throw new XAException(XAException.XAER_INVAL);
        }
        if (xid == rolledback) {
            throw new XAException(XAException.XAER_PROTO);
        }
        try {
            SubordinateCoordinator cor = getCoordinator(xid);
            if (cor.prepareVolatile()) {
                switch (cor.prepare()) {
                    case TwoPhaseOutcome.PREPARE_OK:
                        prepared = xid;
                        return XAResource.XA_OK;
                    case TwoPhaseOutcome.PREPARE_READONLY:
                        return XAResource.XA_RDONLY;
                    case TwoPhaseOutcome.PREPARE_NOTOK:
                    default:
                        return XAResource.XA_RDONLY;
                }
            } else {
                return XAResource.XA_RDONLY;
            }
        } catch (Exception ex) {
            ex.printStackTrace();
            throw new XAException(XAException.XAER_RMFAIL);
        }
    }

    /**
     *
     * @param xid
     * @param onePhase
     * @throws javax.transaction.xa.XAException
     */
    public void commit(Xid xid, boolean onePhase) throws XAException {
        if (xid == null) {
            throw new XAException(XAException.XAER_INVAL);
        }
        if (xid == rolledback) {
            throw new XAException(XAException.XAER_PROTO);
        }
        if (prepared != xid) {
            if (onePhase) {
                prepare(xid);
            } else {
                throw new XAException(XAException.XAER_PROTO);
            }
        }
        try {
            SubordinateCoordinator cor = getCoordinator(xid);
            cor.commit();
            cor.commitVolatile();
            committed = xid;
        } catch (Exception ex) {
            ex.printStackTrace();
            throw new XAException(XAException.XAER_RMFAIL);
        }
    }

    /**
     *
     * @param xid
     * @throws javax.transaction.xa.XAException
     */
    public void rollback(Xid xid) throws XAException {
        if (xid == null) {
            throw new XAException(XAException.XAER_INVAL);
        }
        if (xid == rolledback) {
            throw new XAException(XAException.XAER_PROTO);
        }
        if (xid == committed) {
            throw new XAException(XAException.XAER_PROTO);
        }
        try {
            SubordinateCoordinator cor = getCoordinator(xid);
            cor.rollback();
            cor.rollbackVolatile();
            rolledback = xid;
        } catch (Exception ex) {
            ex.printStackTrace();
            throw new XAException(XAException.XAER_RMFAIL);
        }
    }

    /**
     * 
     * @param xid
     * @throws javax.transaction.xa.XAException
     */
    public void forget(Xid xid) throws XAException {
        if (xid == null) {
            throw new XAException(XAException.XAER_INVAL);
        }
        if (currentXid == null) {
            throw new XAException(XAException.XAER_NOTA);
        }
        currentXid = null;
    }

    /**
     *
     * @param flag
     * @return Successfull xids array
     * @throws javax.transaction.xa.XAException
     */
    public Xid[] recover(int flag) throws XAException {
        return (Xid[]) successfulXids.toArray(new Xid[successfulXids.size()]);
    }

    /**
     *
     * @param xaResource
     * @return Result of the test XAResorce
     */
    public boolean isSameRM(XAResource xaResource) {
        if (!(xaResource instanceof BridgeXAResourceAT)) {
            return false;
        }
        BridgeXAResourceAT other = (BridgeXAResourceAT) xaResource;
        return other.context.equals(this.context);
    }

    /**
     *
     * @return Transaction timeout in mils
     * @throws javax.transaction.xa.XAException
     */
    public int getTransactionTimeout() throws XAException {
        return transactionTimeoutSeconds;
    }

    /**
     *
     * @param seconds
     * @return Result of the set transaction timeouts
     * @throws javax.transaction.xa.XAException
     */
    public boolean setTransactionTimeout(int seconds) throws XAException {
        transactionTimeoutSeconds = seconds;
        return true;
    }

    /**
     *
     * @return Current xid
     */
    public Xid getCurrentXid() {
        return currentXid;
    }

    /**
     *
     */
    public void clear() {
        currentXid = null;
        prepared = null;
        rolledback = null;
        committed = null;
        knownXids.clear();
        successfulXids.clear();
    }

    /**
     *
     * @return Recovered status
     */
    public boolean isRecovered() {
        return recovered;
    }

    /**
     *
     * @param xid 
     * @return Get or recovery or create subordinate transaction coordinator
     * @throws javax.transaction.xa.XAException
     */
    public SubordinateCoordinator getCoordinator(Xid xid) throws XAException {
        try {
            if (!isRecovered()) {
                if (coordinator == null) {
                    coordinator = createSubordinateTransaction();
                    coordinatorId = coordinator.get_uid().stringForm();
                }
                return coordinator;
            } else {
                XTSATRecoveryManager recoveryManager = XTSATRecoveryManager.getRecoveryManager();
                boolean isRecoveryScanStarted = recoveryManager.isSubordinateCoordinatorRecoveryStarted();
                coordinator = SubordinateCoordinator.getRecoveredCoordinator(coordinatorId);
                if (coordinator == null) {
                    if (!isRecoveryScanStarted) {
                        throw new SystemException();
                    }
                } else if (!coordinator.isActivated()) {
                    throw new SystemException();
                }
                return coordinator;
            }
        } catch (Exception ex) {
            throw new XAException(ex.getMessage());
        }
    }

    /**
     *
     * @return Transaction context from subordinate transaction
     * @throws javax.transaction.xa.XAException
     */
    public TxContextImple getTransactionContext() throws XAException {
        if (context == null) {
            ServiceRegistry serviceRegistry = ServiceRegistry.getRegistry();
            String registrationCoordinatorURI = serviceRegistry.getServiceURI(CoordinationConstants.REGISTRATION_SERVICE_NAME);
            CoordinationContext coordinationContext = new CoordinationContext();
            coordinationContext.setCoordinationType(AtomicTransactionConstants.WSAT_PROTOCOL);
            CoordinationContextType.Identifier identifier = new CoordinationContextType.Identifier();
            identifier.setValue("urn:" + coordinatorId);
            coordinationContext.setIdentifier(identifier);

            if (transactionTimeoutSeconds > 0) {
                Expires expiresInstance = new Expires();
                expiresInstance.setValue(transactionTimeoutSeconds);
                coordinationContext.setExpires(expiresInstance);
            }
            W3CEndpointReference registrationCoordinator = getRegistrationCoordinator(registrationCoordinatorURI, coordinatorId);
            coordinationContext.setRegistrationService(registrationCoordinator);
            context = new TxContextImple(new ContextImple(coordinationContext));
        }
        return context;
    }

    private SubordinateCoordinator createSubordinateTransaction() throws NoActivityException, InvalidProtocolException, SystemException {
        final ContextFactoryMapper contextFactoryMapper = ContextFactoryMapper.getFactory();
        final ContextFactory contextFactory = contextFactoryMapper.getContextFactory(AtomicTransactionConstants.WSAT_PROTOCOL);
        final ContextFactoryImple imple = (ContextFactoryImple) contextFactory;
        final SubordinateCoordinator subTx = (SubordinateCoordinator) imple.createSubordinate();
        return subTx;
    }

    private static W3CEndpointReference getRegistrationCoordinator(String registrationCoordinatorURI, String identifier) {
        final W3CEndpointReferenceBuilder builder = new W3CEndpointReferenceBuilder();
        builder.serviceName(CoordinationConstants.REGISTRATION_SERVICE_QNAME);
        builder.endpointName(CoordinationConstants.REGISTRATION_ENDPOINT_QNAME);
        builder.address(registrationCoordinatorURI);
        InstanceIdentifier.setEndpointInstanceIdentifier(builder, identifier);
        W3CEndpointReference registrationCoordinator = builder.build();
        return registrationCoordinator;
    }

    private void readObject(ObjectInputStream aInputStream) throws ClassNotFoundException, IOException {
        aInputStream.defaultReadObject();
        recovered = true;
    }

    private void writeObject(ObjectOutputStream aOutputStream) throws IOException {
        aOutputStream.defaultWriteObject();
    }
}

